#include <stdio.h>
#include <ctype.h>
#include "algebra.h"
#include "cam.h"
#include "globals.h"
#include "scene.h"

void
cam_init(struct cam*c)
{
  struct cam _c=
  {
    {-20.0,0.0,0.0},
    {1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0},
    10.0
  };
  (*c)=_c;
}

void
cam_print(struct cam*c)
{
  fprintf(logfile,"cam -------->\n");
  fprintf(logfile,"self=(%.2f,%.2f,%.2f) r=%.2f\n",
    c->s[0],c->s[1],c->s[2],c->r);
  print_rot(c->a);
}

/* camera position operation */
/* private */
void
_view(struct cam*c,float*x,float*vx)
{
  /*** translate **/
  int i;
  for(i=0;i<3;i++)
    vx[i]=x[i]-c->s[i];
  /** rotate **/
  rotate(c->a,vx,vx);
}

/* project image */
/* private */
void
_project(struct cam*c,float*x,float*px)
{
  /** use c to send x[3] to px[2] ***/
  if(x[0]==0.0)
    /*** if all else fails...****/
    x[0]=epsilon;
  px[0]=-x[1]*c->r/x[0];
  px[1]=x[2]*c->r/x[0];
}

/* private */
/******** should really be in scene.c *****/
void
cam_image_source_item(
  struct cam*c,
  struct image_item*im,
  struct source_item*si,
  int nx)
{
  int j;

  im->depth=0.0;
  im->skip=0;
  /* count through the vertices */
  for(j=0;j<nx;j++)
  {
    float vx[3];
    /* camera viewpoint */
    _view(c,si->x+3*j,vx);

    /* add a depth */
    im->depth+=vx[0];

    if(vx[0]<epsilon)
      im->skip=1; /* can't see this one */

    /** project vertex to image coordinates **/
    _project(c,vx,im->x+2*j);
  }
  im->depth/=(float)nx; /* it's an average now */
}

void
_look_at(
  struct cam*c,
  float*t, /* target */
  float*up /* up vector */ )
{
  int i;
  float r;
  float a[3];
  for(i=0;i<3;i++)
    a[i]=t[i]-c->s[i];
  r=sqrt(dot3(a,a));
  if(r<epsilon)
  {
    fprintf(logfile,"\nlook_at: can't look at self. ignoring.\n");
    return;
  }
  for(i=0;i<3;i++)
    a[i]/=r;
  /* orthogonalize up vector */
  r=dot3(up,a);
  for(i=0;i<3;i++)
    up[i]-=a[i]*r;
  r=sqrt(dot3(up,up));
  if(r<epsilon)
  {
    fprintf(logfile,"\nlook_at: up is collapsed. ignoring.\n");
    return;
  }
  for(i=0;i<3;i++)
    up[i]/=r;
  /* left = up cross a */
  c->a[3]=up[1]*a[2]-up[2]*a[1];
  c->a[4]=up[2]*a[0]-up[0]*a[2];
  c->a[5]=up[0]*a[1]-up[1]*a[0];
  for(i=0;i<3;i++)
  {
    c->a[i]=a[i];
    c->a[i+6]=up[i];
  }
}

void
cam_center(struct cam*c,int bi,int bj,int bk)
{
  float t[3]; 
  float up[3]={0.0, 0.0, 1.0};
  t[0]=bi/2.0;
  t[1]=bj/2.0;
  t[2]=bk/2.0;
  _look_at(c,t,up);
}

void
cam_up(struct cam*c)
{
  float r[9];
  y_rot(r,-0.1);
  mult(c->a,r,c->a);
}

void
cam_down(struct cam*c)
{
  float r[9];
  y_rot(r,0.1);
  mult(c->a,r,c->a);
}

void
cam_left(struct cam*c)
{
  float r[9];
  z_rot(r,-0.1);
  mult(c->a,r,c->a);
}

void
cam_right(struct cam*c)
{
  float r[9];
  z_rot(r,0.1);
  mult(c->a,r,c->a);
}

void
cam_in(struct cam*c)
{
  add(c->s,c->a,c->s);
}

void
cam_out(struct cam*c)
{
  minus(c->s,c->a,c->s);
}

void
cam_rotate_up(struct cam*c)
{
  float r[9];
  y_rot(r,0.1);
  rotate(r,c->s,c->s);
  y_rot(r,-0.1);
  mult(c->a,r,c->a);
}

void
cam_rotate_down(struct cam*c)
{
  float r[9];
  y_rot(r,-0.1);
  rotate(r,c->s,c->s);
  y_rot(r,0.1);
  mult(c->a,r,c->a);
}

void
cam_rotate_left(struct cam*c)
{
  float r[9];
  z_rot(r,0.1);
  rotate(r,c->s,c->s);
  z_rot(r,-0.1);
  mult(c->a,r,c->a);
}

void
cam_rotate_right(struct cam*c)
{
  float r[9];
  z_rot(r,-0.1);
  rotate(r,c->s,c->s);
  z_rot(r,0.1);
  mult(c->a,r,c->a);
}

void
cam_flip(struct cam*c,float x[3])
{
  minus(c->s,x,c->s);
  c->s[0]=-c->s[0];
  c->s[1]=-c->s[1];
  c->s[2]=-c->s[2];
  c->a[0]=-c->a[0];
  c->a[3]=-c->a[3];
  c->a[6]=-c->a[6];
  c->a[1]=-c->a[1];
  c->a[4]=-c->a[4];
  c->a[7]=-c->a[7];
  add(c->s,x,c->s);
}

void
cam_track_up(struct cam*c,float x[3],float deg)
{
  float r[9],_r[9];
  float theta=3.1416*deg/180.0;
  minus(c->s,x,c->s);
  y_rot(r,-theta);
  trans(c->a,_r);
  mult(r,c->a,r);
  mult(_r,r,r);
  rotate(r,c->s,c->s);
  trans(c->a,_r);
  mult(r,_r,_r);
  trans(_r,c->a);
  ortho(c->a); /* straighten out */
  add(c->s,x,c->s);
}

void
cam_track_down(struct cam*c,float x[3], float deg)
{
  float r[9],_r[9];
  float theta=3.1416*deg/180.0;
  minus(c->s,x,c->s);
  y_rot(r,theta);
  trans(c->a,_r);
  mult(r,c->a,r);
  mult(_r,r,r);
  rotate(r,c->s,c->s);
  trans(c->a,_r);
  mult(r,_r,_r);
  trans(_r,c->a);
  ortho(c->a); /* straighten out */
  add(c->s,x,c->s);
}

void
cam_track_left(struct cam*c,float x[3], float deg)
{
  float r[9],_r[9];
  float theta=3.1416*deg/180.0;
  minus(c->s,x,c->s);
  z_rot(r,theta);
  trans(c->a,_r);
  mult(r,c->a,r);
  mult(_r,r,r);
  rotate(r,c->s,c->s);
  trans(c->a,_r);
  mult(r,_r,_r);
  trans(_r,c->a);
  ortho(c->a); /* straighten out */
  add(c->s,x,c->s);
}

void
cam_track_right(struct cam*c,float x[3], float deg)
{
  float r[9],_r[9];
  float theta=3.1416*deg/180.0;
  minus(c->s,x,c->s);
  z_rot(r,-theta);
  trans(c->a,_r);
  mult(r,c->a,r);
  mult(_r,r,r);
  rotate(r,c->s,c->s);
  trans(c->a,_r);
  mult(r,_r,_r);
  trans(_r,c->a);
  ortho(c->a); /* straighten out */
  add(c->s,x,c->s);
}

void
cam_zoom_in(struct cam*c)
{
  c->r*=1.1;
}

void
cam_zoom_out(struct cam*c)
{
  c->r*=0.9;
}

