/* map.h */
#ifndef __MAP_H
#define __MAP_H

#include <unistd.h>
#include <sys/time.h>
#include "ivec.h"

struct cam;
struct scene;
struct draw_item;

struct boundary
{
  struct draw_item*d;
  char refcount;
};

#define BCL_SZ 6

struct bcluster
{
  /** some relevant pointer ... */
  void*owner;
  /** the center of the cluster */
  int iface, i0, i1;
  /** it's shape */
  char mask[BCL_SZ][BCL_SZ][2];
  /** linked list */
  struct bcluster*next;
};

struct hole
{
  /** where */
  int i,j,k; 
  /** when it will fill */
#ifdef USEC
  struct timeval tv;
#else
  int msec;
#endif

  /** linked list */
  struct hole*next;
};

#define HOLE_FRAMES 70

struct parcel
{
  /** where */
  int i,j,k;
  struct draw_item*d;
  struct parcel*next;
};

struct mapel
{
  unsigned char b; /** content */
  unsigned char l; /** lock count */
  unsigned char u[6][6]; /** robot goal info */
};

struct map
{
  struct scene*s;
  int bi,bj,bk;

  /** all the bricks, el[i][j][k] */
  struct mapel*el;

  /** squares: sq[i][j][k][l]; for the bricks */
  struct draw_item** sq; 

  /** list of parcels */
  struct parcel*pa; /* sentinel */
  int n_parcels; /* total number of parcels */

  /** list of holes */
  struct hole*ho; /* sentinel */

  /** boundary indicators, bd[6][][] */
  struct boundary** bd[6];

  /** some helpful arrays */
  int fsz[3];
  int bsz[3];
  int brg[3][2];

  /** sentinel */
  struct bcluster*bcl;

  /** portal **/
  struct ivec*piv;
  int portal_on_sq; /** redundant ... */
};

/* TODO: implement sparse maps ala. Knuth 2.2.6 */

/** defines for mapel.b */
#define EMPTY_EL	0
#define BRICK_EL	1
#define HOLE_EL		(1<<1)
#define PARCEL_EL	(1<<2)
#define PORTAL_EL	(1<<3)
#define ROBOT_EL	(1<<4)
#define SOLID_EL	(1<<5)

void
map_init(struct map*m,struct scene*s,int bx,int by,int bz);

void
map_compile(struct map*);

void
map_notify_object(struct map*,void*,struct ivec);

void
map_notify_object_removal(struct map*,void*);

void
map_cull_backfaces(struct map*,struct cam*);

/** how do we get from s_iv to t_iv ? */
struct ivec
map_query(
  struct map*m,
  struct ivec s_iv,
  struct ivec t_iv );

int
map_in_range(struct map*m,int i,int j,int k);

#ifdef USEC
void
map_poll(struct map*m,struct timeval tv);
#else
void
map_poll(struct map*m,int msec);
#endif

/********** hole **************/

int
map_is_hole(struct map*m,int i, int j,int k);

#ifdef USEC
void
map_set_hole(
  struct map*m,int i, int j, int k,struct timeval tv);
#else
void
map_set_hole(
  struct map*m,int i, int j, int k,int msec);
#endif

/******** lock *******************/

int
map_is_lock(struct map*m,int i, int j, int k);

void
map_set_lock(struct map*m,int i, int j, int k);

void
map_unset_lock(struct map*m,int i, int j, int k);

/********** brick ************/

int
map_is_brick(struct map*m,int i, int j, int k);

int
map_is_solid(struct map*m,int i, int j, int k);

void
map_set_brick(struct map*m,int i,int j,int k,int solid);

void
map_unset_brick(struct map*m,int i,int j,int k);

/********** parcel *************/

void
map_set_parcel(struct map*m,int i,int j,int k);

int
map_is_parcel(struct map*m,int i, int j, int k);

void
map_unset_parcel(struct map*m,int i,int j,int k);

int
map_n_parcels(struct map*);

void
map_toggle_parcel(struct map*m,int i, int j, int k);

/************ portal ********/

void
map_set_portal(struct map*m,struct ivec iv);

void
map_unset_portal(struct map*m,struct ivec iv);

void
map_toggle_portal(struct map*m,struct ivec);

int
map_is_portal(struct map*m,struct ivec);

int
map_is_parcel(struct map*m,int i,int j,int k);

/***********************************/

float*
map_get_center(struct map*);

void
map_print(struct map*m);

void
map_finalize(struct map*m);

#endif


